from __future__ import annotations

import logging
import secrets
from ipaddress import IPv4Address

import pytest
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey

from aiostem.exceptions import CryptographyError, ReplyStatusError
from aiostem.reply import (
    ReplyAddOnion,
    ReplyAuthChallenge,
    ReplyAuthenticate,
    ReplyExtendCircuit,
    ReplyGetConf,
    ReplyGetInfo,
    ReplyMapAddress,
    ReplyOnionClientAuthView,
    ReplyProtocolInfo,
)
from aiostem.structures import AuthMethod, OnionServiceKeyStruct, OnionServiceKeyType
from aiostem.utils import Message, messages_from_stream

from .utils.test_message import create_stream

# All test coroutines will be treated as marked for asyncio.
pytestmark = pytest.mark.asyncio


async def create_message(lines: list[str]) -> Message:
    """
    Get a single message from the provided lines.

    Args:
        lines: a list of lines to build a stream from.

    Raises:
        RuntimeError: when no message was found.

    Returns:
        The first message extracted from the lines.

    """
    stream = create_stream(lines)
    async for msg in messages_from_stream(stream):
        return msg
    msg = 'Unable to find a message in the provided lines.'
    raise RuntimeError(msg)


class TestReplies:
    """Check that replies are properly parsed."""

    async def test_get_conf(self):
        lines = [
            '250-ControlPort=0.0.0.0:9051',
            '250-Log=notice stdout',
            '250-EntryNode=',
            '250 HashedControlPassword',
        ]
        message = await create_message(lines)
        reply = ReplyGetConf.from_message(message)
        assert len(reply) == 4
        assert reply['ControlPort'] == '0.0.0.0:9051'
        assert reply['Log'] == 'notice stdout'
        assert reply['EntryNode'] == ''
        assert reply['HashedControlPassword'] is None
        assert set(reply.values()) == {'0.0.0.0:9051', 'notice stdout', '', None}

    async def test_get_conf_empty(self):
        message = await create_message(['250 OK'])
        reply = ReplyGetConf.from_message(message)
        assert len(reply) == 0
        assert reply.status_text == 'OK'
        assert reply.is_success is True

    async def test_get_conf_error(self):
        lines = ['552 Unrecognized configuration key "A"']
        message = await create_message(lines)
        reply = ReplyGetConf.from_message(message)
        with pytest.raises(ReplyStatusError, match='Unrecognized configuration key "A"'):
            reply.raise_for_status()

    async def test_get_conf_multi(self):
        lines = [
            '250-ControlPort=0.0.0.0:9051',
            '250-ControlPort=0.0.0.0:9052',
            '250 ControlPort=0.0.0.0:9053',
        ]
        message = await create_message(lines)
        reply = ReplyGetConf.from_message(message)
        assert set(reply.keys()) == {'ControlPort'}
        assert isinstance(reply['ControlPort'], list)
        assert len(reply['ControlPort']) == 3

    async def test_map_address(self):
        lines = [
            '250-127.218.108.43=bogus1.google.com',
            '250 one.one.one.one=1.1.1.1',
        ]
        message = await create_message(lines)
        reply = ReplyMapAddress.from_message(message)

        assert reply.status == 250
        assert reply.status_text is None

        assert len(reply.items) == 2

        data = reply.items[0].data
        assert data is not None
        assert data.original == IPv4Address('127.218.108.43')
        assert data.replacement == 'bogus1.google.com'

        data = reply.items[1].data
        assert data is not None
        assert data.original == 'one.one.one.one'
        assert data.replacement == IPv4Address('1.1.1.1')

    async def test_map_address_error(self):
        lines = [
            "512-syntax error: invalid address '@@@'",
            '250 one.one.one.one=1.1.1.1',
        ]
        message = await create_message(lines)
        reply = ReplyMapAddress.from_message(message)
        assert reply.status == 512
        assert reply.status_text == "syntax error: invalid address '@@@'"
        assert len(reply.items) == 2

        item = reply.items[0]
        assert item.status == 512
        assert item.data is None

        item = reply.items[1]
        assert item.status == 250
        assert item.data is not None
        assert item.data.original == 'one.one.one.one'
        assert item.data.replacement == IPv4Address('1.1.1.1')

    async def test_get_info(self):
        lines = [
            '250-version=0.4.8.12',
            '250+orconn-status=',
            '$4D0F2ADB9CD55C3EBD14823D54B6541B99A51C19~Unnamed CONNECTED',
            '$DCD645A9C7183A893AC4EF0369AAB5ED1ADBD2AF~fifo4ka CONNECTED',
            '$DAC825BBF05D678ABDEA1C3086E8D99CF0BBF112~malene CONNECTED',
            '$CB44E8ED1FAB648275C39756EB1758060C43BCA4~NOTaGlowieRelay CONNECTED',
            '.',
            '250 OK',
        ]
        message = await create_message(lines)
        reply = ReplyGetInfo.from_message(message)
        assert len(reply) == 2
        assert set(reply) == {'version', 'orconn-status'}

    async def test_get_info_error(self):
        lines = ['552 Not running in server mode']
        message = await create_message(lines)
        reply = ReplyGetInfo.from_message(message)
        assert reply.is_error is True
        assert reply.is_success is False
        assert len(reply) == 0

    async def test_extend_circuit(self):
        lines = ['250 EXTENDED 56832']
        message = await create_message(lines)
        reply = ReplyExtendCircuit.from_message(message)
        reply.raise_for_status()

        assert reply.data.circuit == 56832

    async def test_extend_circuit_error(self):
        lines = ['552 Unknown circuit "12"']
        message = await create_message(lines)
        reply = ReplyExtendCircuit.from_message(message)
        assert reply.status == 552
        assert reply.status_text == 'Unknown circuit "12"'
        assert reply.data is None

    async def test_authenticate(self):
        lines = ['250 OK']
        message = await create_message(lines)
        reply = ReplyAuthenticate.from_message(message)
        assert reply.status == 250
        assert reply.status_text == 'OK'

    async def test_protocol_info(self):
        lines = [
            '250-PROTOCOLINFO 1',
            (
                '250-AUTH METHODS=COOKIE,SAFECOOKIE,HASHEDPASSWORD '
                'COOKIEFILE="/run/tor/control.authcookie"'
            ),
            '250-VERSION Tor="0.4.8.12"',
            '250 OK',
        ]
        message = await create_message(lines)
        reply = ReplyProtocolInfo.from_message(message)
        # Is not supposed to raise anything.
        reply.raise_for_status()

        assert reply.data.protocol_version == 1
        assert reply.data.tor_version == '0.4.8.12'
        assert AuthMethod.COOKIE in reply.data.auth_methods

    async def test_protocol_info_read_cookie(self, tmp_path):
        cookie_data = secrets.token_bytes(32)
        cookie_path = tmp_path / 'cookiefile'
        lines = [
            '250-PROTOCOLINFO 1',
            f'250-AUTH METHODS=COOKIE,SAFECOOKIE COOKIEFILE="{cookie_path}"',
            '250-VERSION Tor="0.4.8.12"',
            '250 OK',
        ]

        with open(cookie_path, 'wb') as fp:  # noqa: ASYNC230
            fp.write(cookie_data)

        message = await create_message(lines)
        reply = ReplyProtocolInfo.from_message(message)
        reply.raise_for_status()

        cookie = await reply.read_cookie_file()
        assert cookie == cookie_data

    async def test_protocol_info_no_cookie_file(self):
        lines = [
            '250-PROTOCOLINFO 1',
            '250-AUTH METHODS=HASHEDPASSWORD',
            '250-VERSION Tor="0.4.8.13"',
            '250 OK',
        ]
        message = await create_message(lines)
        reply = ReplyProtocolInfo.from_message(message)
        reply.raise_for_status()

        assert reply.data.auth_cookie_file is None
        with pytest.raises(FileNotFoundError, match='No cookie file found'):
            await reply.read_cookie_file()

    async def test_protocol_info_error(self):
        message = await create_message(['513 No such version "aa"'])
        reply = ReplyProtocolInfo.from_message(message)
        assert reply.status == 513
        with pytest.raises(ReplyStatusError, match='No such version "aa"'):
            reply.raise_for_status()

    async def test_protocol_info_unknown_line(self, caplog):
        lines = [
            '250-PROTOCOLINFO 1',
            '250-TEST For="science"',
            '250-VERSION Tor="0.4.8.12"',
            '250 OK',
        ]
        message = await create_message(lines)
        with caplog.at_level(logging.INFO, logger='aiostem'):
            ReplyProtocolInfo.from_message(message)
        assert "No syntax handler for keyword 'TEST'" in caplog.text

    async def test_auth_challenge_parse(self):
        server_hash = '700912005E616BC5558ACDC14B11304B0A03F45C4B1DBD60365FD66033D7276C'
        server_nonce = 'E0BEAB4467F69317B5BEE45B04565F5FF277B896A2DE46A86C99B3999F77CE80'
        line = f'250 AUTHCHALLENGE SERVERHASH={server_hash} SERVERNONCE={server_nonce}'
        message = await create_message([line])
        reply = ReplyAuthChallenge.from_message(message)
        assert reply.status == 250
        assert reply.status_text is None

        assert reply.data is not None
        assert reply.data.server_hash == bytes.fromhex(server_hash)
        assert reply.data.server_nonce == bytes.fromhex(server_nonce)

        cookie = b'e8a05005deb487f5d9a0db9a026d28ad'
        with pytest.raises(CryptographyError, match='No client_nonce was found or provided'):
            reply.data.build_client_hash(cookie)
        with pytest.raises(CryptographyError, match='No client_nonce was found or provided'):
            reply.data.build_server_hash(cookie)

    async def test_auth_challenge_syntax_error(self):
        line = '512 Wrong number of arguments for AUTHCHALLENGE'
        message = await create_message([line])
        reply = ReplyAuthChallenge.from_message(message)
        assert reply.status == 512
        assert reply.status_text == 'Wrong number of arguments for AUTHCHALLENGE'
        with pytest.raises(ReplyStatusError, match='Wrong number of arguments'):
            reply.raise_for_status()

    async def test_auth_challenge_error(self):
        client_nonce_str = 'F1BE0456FB2626512D72B06509A16EAAA707B1981F31C9BBAD40A788A0A330A6'
        server_hash = '000912005E616BC5558ACDC14B11304B0A03F45C4B1DBD60365FD66033D7276C'
        server_nonce = 'E0BEAB4467F69317B5BEE45B04565F5FF277B896A2DE46A86C99B3999F77CE80'
        cookie_str = 'E9FDE075EA5C9996F17AB280B3FD69FF6109ECA9369ED824045E8333DB58017A'
        client_nonce = bytes.fromhex(client_nonce_str)
        cookie = bytes.fromhex(cookie_str)

        line = f'250 AUTHCHALLENGE SERVERHASH={server_hash} SERVERNONCE={server_nonce}'
        message = await create_message([line])
        reply = ReplyAuthChallenge.from_message(message)
        with pytest.raises(CryptographyError, match='Server hash provided by Tor is invalid'):
            reply.data.raise_for_server_hash_error(cookie, client_nonce)

    async def test_auth_challenge_success_bytes(self):
        client_nonce_str = 'F1BE0456FB2626512D72B06509A16EAAA707B1981F31C9BBAD40A788A0A330A6'
        server_hash = '700912005E616BC5558ACDC14B11304B0A03F45C4B1DBD60365FD66033D7276C'
        server_nonce = 'E0BEAB4467F69317B5BEE45B04565F5FF277B896A2DE46A86C99B3999F77CE80'
        cookie_str = 'E9FDE075EA5C9996F17AB280B3FD69FF6109ECA9369ED824045E8333DB58017A'
        client_nonce = bytes.fromhex(client_nonce_str)
        cookie = bytes.fromhex(cookie_str)

        line = f'250 AUTHCHALLENGE SERVERHASH={server_hash} SERVERNONCE={server_nonce}'
        message = await create_message([line])
        reply = ReplyAuthChallenge.from_message(message)
        reply.data.raise_for_server_hash_error(cookie, client_nonce)

        client_hash_str = 'DDC1E1FC978DDF6CD2142EEA62559D026A2F84666B9F6B462224F36B7E9A9C54'
        client_hash = bytes.fromhex(client_hash_str)
        computed = reply.data.build_client_hash(cookie, client_nonce)
        assert computed == client_hash

    async def test_auth_challenge_success_string(self):
        server_hash = '01BAB534C249A47B46D8CA235683B43D075C134820CAF3C0214DBDE2ADD55ED3'
        server_nonce = '76A744F700967CE08FE7E45797FA54BDCDEBF12F273080EB58562B80DBD02400'
        cookie_str = 'E9FDE075EA5C9996F17AB280B3FD69FF6109ECA9369ED824045E8333DB58017A'
        cookie = bytes.fromhex(cookie_str)
        client_nonce = 'I am a nonce!'

        line = f'250 AUTHCHALLENGE SERVERHASH={server_hash} SERVERNONCE={server_nonce}'
        message = await create_message([line])
        reply = ReplyAuthChallenge.from_message(message)
        reply.data.raise_for_server_hash_error(cookie, client_nonce)

        client_hash_str = '2E1DA1886E1D4D2695F10290C315877E55D838DAA04757E6D9730420DD39262C'
        client_hash = bytes.fromhex(client_hash_str)
        computed = reply.data.build_client_hash(cookie, client_nonce)
        assert computed == client_hash

    async def test_add_onion(self):
        lines = [
            '250-ServiceID=xsa5oiu2bpnpvsgec5ti5gqdue4t4uxfmmobqibj5nuwkzxb2krtlgyd',
            '250-PrivateKey=ED25519-V3:wCuCqxopubG0hu7WyeWsZvTa7ipqRCAnIVnghhc0pFvP'
            '6CYUzsZJCNw3bBcuita8Dr59xaUqM2nJBFZRthLTtw==',
            '250 OK',
        ]
        message = await create_message(lines)
        reply = ReplyAddOnion.from_message(message)
        reply.raise_for_status()

        data = reply.data
        assert data is not None
        assert isinstance(data.key, OnionServiceKeyStruct)
        assert data.key.key_type == OnionServiceKeyType.ED25519_V3
        assert data.key.data.hex() == (
            'c02b82ab1a29b9b1b486eed6c9e5ac66f4daee2a6a4420272159e0861734a45b'
            'cfe82614cec64908dc376c172e8ad6bc0ebe7dc5a52a3369c9045651b612d3b7'
        )

    async def test_add_onion_with_service_key(self):
        lines = [
            '250-ServiceID=xsa5oiu2bpnpvsgec5ti5gqdue4t4uxfmmobqibj5nuwkzxb2krtlgyd',
            '250 OK',
        ]
        message = await create_message(lines)
        reply = ReplyAddOnion.from_message(message)
        assert reply.data is not None
        assert reply.data.key is None

    async def test_add_onion_with_client_auth_v3(self):
        lines = [
            '250-ServiceID=xin2djtit46y642io4deov5r6z63f2gz2ijpawxkrvm4kr3kjsu744ad',
            '250-PrivateKey=ED25519-V3:sEOmgu/tc8jCkp58dkXZho66OjhDwvwc48qkQ2CeAlbE'
            'zGnm22+kXNBH+rN5MQYSHWCMmYTjMbxbS7KAVSM+Yw==',
            '250-ClientAuthV3=5BPBXQOAZWPSSXFKOIXHZDRDA2AJT2SWS2GIQTISCFKGVBFWBBDQ',
            '250-ClientAuthV3=RC3BHJ6WTBQPRRSMV65XGCZVSYJQZNWBQI3LLFS73VP6NHSIAD2Q',
            '250 OK',
        ]
        message = await create_message(lines)
        reply = ReplyAddOnion.from_message(message)
        assert len(reply.data.client_auth_v3) == 2
        for client_auth in reply.data.client_auth_v3:
            assert isinstance(client_auth, X25519PublicKey)

    async def test_add_onion_error(self):
        line = '512 Bad arguments to ADD_ONION: Need at least 1 argument(s)'
        message = await create_message([line])
        reply = ReplyAddOnion.from_message(message)
        with pytest.raises(ReplyStatusError, match='Bad arguments to ADD_ONION:'):
            reply.raise_for_status()

    async def test_onion_client_auth_view(self):
        lines = [
            (
                '250-ONION_CLIENT_AUTH_VIEW '
                'aiostem26gcjyybsi3tyek6txlivvlc5tczytz52h4srsttknvd5s3qd'
            ),
            (
                '250-CLIENT aiostem26gcjyybsi3tyek6txlivvlc5tczytz52h4srsttknvd5s3qd '
                'x25519:yPGUxgKaC5ACyEzsdANHJEJzt5DIqDRBlAFaAWWQn0o= ClientName=Peter'
            ),
            '250 OK',
        ]
        message = await create_message(lines)
        reply = ReplyOnionClientAuthView.from_message(message)

        assert reply.data is not None
        assert len(reply.data.clients) == 1

        client = reply.data.clients[0]
        assert client.address == 'aiostem26gcjyybsi3tyek6txlivvlc5tczytz52h4srsttknvd5s3qd'
        assert client.name == 'Peter'
        assert isinstance(client.key, X25519PrivateKey)
        assert len(client.flags) == 0

    async def test_onion_client_auth_view_error(self):
        line = '512 Invalid v3 address "hjg"'
        message = await create_message([line])
        reply = ReplyOnionClientAuthView.from_message(message)
        with pytest.raises(ReplyStatusError, match='Invalid v3 address "hjg"'):
            reply.raise_for_status()



from __future__ import annotations

import secrets
from base64 import b32decode, b64decode, b64encode

import pytest
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from pydantic import TypeAdapter
from pydantic_core import PydanticSerializationError

from aiostem.command import (
    CommandAddOnion,
    CommandAttachStream,
    CommandAuthChallenge,
    CommandAuthenticate,
    CommandCloseCircuit,
    CommandCloseStream,
    CommandDelOnion,
    CommandDropGuards,
    CommandDropOwnership,
    CommandDropTimeouts,
    CommandExtendCircuit,
    CommandGetConf,
    CommandGetInfo,
    CommandHsFetch,
    CommandHsPost,
    CommandLoadConf,
    CommandMapAddress,
    CommandOnionClientAuthAdd,
    CommandOnionClientAuthRemove,
    CommandOnionClientAuthView,
    CommandPostDescriptor,
    CommandProtocolInfo,
    CommandQuit,
    CommandRedirectStream,
    CommandResetConf,
    CommandResolve,
    CommandSaveConf,
    CommandSerializer,
    CommandSetCircuitPurpose,
    CommandSetConf,
    CommandSetEvents,
    CommandSignal,
    CommandTakeOwnership,
    CommandUseFeature,
    CommandWord,
)
from aiostem.event import EventWord
from aiostem.exceptions import CommandError
from aiostem.structures import (
    CircuitPurpose,
    DescriptorPurpose,
    HsDescAuthCookie,
    HsDescAuthTypeInt,
    HsDescClientAuthV2,
    LongServerName,
    OnionClientAuthFlags,
    OnionServiceFlags,
    OnionServiceKeyStruct,
    OnionServiceKeyType,
    OnionServiceNewKeyStruct,
    Signal,
    StreamCloseReasonInt,
    VirtualPort,
)
from aiostem.utils import ArgumentKeyword, ArgumentString, QuoteStyle, TrEd25519PrivateKey

VirtualPortAdapter = TypeAdapter(VirtualPort)


class TestCommandSerializer:
    """Check that the command serializer works."""

    def test_default_properties(self):
        ser = CommandSerializer(CommandWord.SETCONF)
        assert ser.command == CommandWord.SETCONF
        assert len(ser.arguments) == 0
        assert ser.body is None

    def test_serialize_argument(self):
        ser = CommandSerializer(CommandWord.SETCONF)
        arg = ArgumentKeyword(None, 'hello', quotes=QuoteStyle.ALWAYS)
        ser.arguments.append(arg)
        assert len(ser.arguments) == 1
        assert ser.serialize() == 'SETCONF "hello"\r\n'

    def test_serialize_simple_body(self):
        ser = CommandSerializer(CommandWord.SETCONF)
        ser.body = 'Hello world'
        assert ser.serialize() == '+SETCONF\r\nHello world\r\n.\r\n'

    def test_serialize_multiline_body(self):
        ser = CommandSerializer(CommandWord.SETCONF)
        ser.body = 'Hello world\n.dot'
        assert ser.serialize() == '+SETCONF\r\nHello world\r\n..dot\r\n.\r\n'

    def test_line_injection(self):
        ser = CommandSerializer(CommandWord.SETCONF)
        ser.arguments.append(ArgumentString('\r\nQUIT'))
        with pytest.raises(CommandError, match='Command injection was detected'):
            ser.serialize()


class TestCommands:
    """Test all commands."""

    def test_set_conf_with_value(self):
        cmd = CommandSetConf(values={'ControlPort': 9872})
        assert cmd.serialize() == 'SETCONF ControlPort=9872\r\n'

    def test_set_conf_with_list(self):
        cmd = CommandSetConf(values={'ControlPort': [9872, 1234]})
        assert cmd.serialize() == 'SETCONF ControlPort=9872 ControlPort=1234\r\n'

    def test_set_conf_with_null(self):
        cmd = CommandSetConf(values={'ControlPort': None})
        assert cmd.serialize() == 'SETCONF ControlPort\r\n'

    def test_set_conf_error(self):
        with pytest.raises(PydanticSerializationError, match='No value provided'):
            CommandSetConf().serialize()

    def test_reset_conf_with_value(self):
        cmd = CommandResetConf(values={'ControlPort': 9872})
        assert cmd.serialize() == 'RESETCONF ControlPort=9872\r\n'

    def test_reset_conf_with_null(self):
        cmd = CommandResetConf(values={'ControlPort': None})
        assert cmd.serialize() == 'RESETCONF ControlPort\r\n'

    def test_reset_conf_error(self):
        with pytest.raises(PydanticSerializationError, match='No value provided'):
            CommandResetConf().serialize()

    def test_get_conf(self):
        cmd = CommandGetConf(keywords=['ControlPort', 'PIDFile'])
        assert cmd.serialize() == 'GETCONF ControlPort PIDFile\r\n'

    def test_set_events_circ(self):
        cmd = CommandSetEvents(events={EventWord.CIRC})
        assert cmd.serialize() == 'SETEVENTS CIRC\r\n'

    def test_authenticate_with_password(self):
        cmd = CommandAuthenticate(token='A real stuff')  # noqa: S106
        assert cmd.serialize() == 'AUTHENTICATE "A real stuff"\r\n'

    def test_authenticate_with_token(self):
        token = b'A real stuff'
        cmd = CommandAuthenticate(token=token)
        assert cmd.serialize() == f'AUTHENTICATE {token.hex()}\r\n'

    def test_authenticate_with_null(self):
        cmd = CommandAuthenticate(token=None)
        assert cmd.serialize() == 'AUTHENTICATE\r\n'

    def test_save_conf_standard(self):
        cmd = CommandSaveConf()
        assert cmd.serialize() == 'SAVECONF\r\n'

    def test_save_conf_forced(self):
        cmd = CommandSaveConf(force=True)
        assert cmd.serialize() == 'SAVECONF FORCE\r\n'

    def test_signal(self):
        cmd = CommandSignal(signal=Signal.NEWNYM)
        assert cmd.serialize() == 'SIGNAL NEWNYM\r\n'

    def test_map_address(self):
        cmd = CommandMapAddress(addresses={'1.2.3.4': 'torproject.org'})
        assert cmd.serialize() == 'MAPADDRESS 1.2.3.4=torproject.org\r\n'

    def test_map_address_error(self):
        with pytest.raises(PydanticSerializationError, match='No address provided'):
            CommandMapAddress().serialize()

    def test_get_info(self):
        cmd = CommandGetInfo(keywords=['version', 'config-file'])
        assert cmd.serialize() == 'GETINFO version config-file\r\n'

    def test_get_info_error(self):
        with pytest.raises(PydanticSerializationError, match='No keyword provided'):
            CommandGetInfo().serialize()

    def test_extend_circuit_simple(self):
        cmd = CommandExtendCircuit(circuit=0)
        assert cmd.serialize() == 'EXTENDCIRCUIT 0\r\n'

    def test_extend_circuit_advanced(self):
        cmd = CommandExtendCircuit(
            circuit=12345,
            servers=[
                LongServerName.from_string('$b34a4ac3892e41c58709d9c51b3648620a7d5bfe~Test1'),
                LongServerName.from_string('$7b70bf914770f022e71a26cbf3d9519dc89f2a9a~Test2'),
            ],
            purpose=CircuitPurpose.GENERAL,
        )
        assert cmd.serialize() == (
            'EXTENDCIRCUIT '
            '12345 '
            '$B34A4AC3892E41C58709D9C51B3648620A7D5BFE~Test1,'
            '$7B70BF914770F022E71A26CBF3D9519DC89F2A9A~Test2 '
            'purpose=GENERAL'
            '\r\n'
        )

    def test_set_circuit_purpose(self):
        cmd = CommandSetCircuitPurpose(circuit=0, purpose=CircuitPurpose.CONTROLLER)
        assert cmd.serialize() == 'SETCIRCUITPURPOSE 0 purpose=CONTROLLER\r\n'

    def test_attach_stream(self):
        cmd = CommandAttachStream(circuit=12, stream=2134)
        assert cmd.serialize() == 'ATTACHSTREAM 2134 12\r\n'

    def test_attach_stream_with_hop(self):
        cmd = CommandAttachStream(circuit=12, stream=2134, hop=5)
        assert cmd.serialize() == 'ATTACHSTREAM 2134 12 HOP=5\r\n'

    def test_post_descriptor(self):
        cmd = CommandPostDescriptor(descriptor='This is a descriptor')
        assert cmd.serialize() == '+POSTDESCRIPTOR\r\nThis is a descriptor\r\n.\r\n'

    def test_post_descriptor_advanced(self):
        cmd = CommandPostDescriptor(
            cache=True,
            descriptor='desc',
            purpose=DescriptorPurpose.GENERAL,
        )
        assert cmd.serialize() == '+POSTDESCRIPTOR purpose=general cache=yes\r\ndesc\r\n.\r\n'

    def test_redirect_stream(self):
        cmd = CommandRedirectStream(stream=1234, address='127.0.0.1')
        assert cmd.serialize() == 'REDIRECTSTREAM 1234 127.0.0.1\r\n'

    def test_redirect_stream_with_port(self):
        cmd = CommandRedirectStream(stream=1234, address='127.0.0.1', port=8443)
        assert cmd.serialize() == 'REDIRECTSTREAM 1234 127.0.0.1 8443\r\n'

    def test_close_stream(self):
        cmd = CommandCloseStream(stream=1234, reason=StreamCloseReasonInt.TIMEOUT)
        assert cmd.serialize() == 'CLOSESTREAM 1234 7\r\n'

    def test_close_circuit(self):
        cmd = CommandCloseCircuit(circuit=1234)
        assert cmd.serialize() == 'CLOSECIRCUIT 1234\r\n'

    def test_close_circuit_with_flags(self):
        cmd = CommandCloseCircuit(circuit=1234, if_unused=True)
        assert cmd.serialize() == 'CLOSECIRCUIT 1234 IfUnused\r\n'

    def test_quit(self):
        cmd = CommandQuit()
        assert cmd.serialize() == 'QUIT\r\n'

    def test_use_feature(self):
        cmd = CommandUseFeature(features={'VERBOSE_NAMES'})
        assert cmd.serialize() == 'USEFEATURE VERBOSE_NAMES\r\n'

    def test_resolve(self):
        cmd = CommandResolve(addresses=['torproject.org'])
        assert cmd.serialize() == 'RESOLVE torproject.org\r\n'

    def test_resolve_reverse(self):
        cmd = CommandResolve(addresses=['1.1.1.1'], reverse=True)
        assert cmd.serialize() == 'RESOLVE mode=reverse 1.1.1.1\r\n'

    def test_protocol_info(self):
        cmd = CommandProtocolInfo()
        assert cmd.serialize() == 'PROTOCOLINFO\r\n'

    def test_protocol_info_with_version(self):
        cmd = CommandProtocolInfo(version=1)
        assert cmd.serialize() == 'PROTOCOLINFO 1\r\n'

    def test_load_conf(self):
        cmd = CommandLoadConf(text='SocksPort 127.0.0.1:9050\n')
        assert cmd.serialize() == '+LOADCONF\r\nSocksPort 127.0.0.1:9050\r\n.\r\n'

    def test_take_ownership(self):
        cmd = CommandTakeOwnership()
        assert cmd.serialize() == 'TAKEOWNERSHIP\r\n'

    def test_auth_challenge_bytes(self):
        nonce = secrets.token_bytes(CommandAuthChallenge.NONCE_LENGTH)
        cmd = CommandAuthChallenge(nonce=nonce)
        assert cmd.serialize() == f'AUTHCHALLENGE SAFECOOKIE {nonce.hex()}\r\n'

    def test_auth_challenge_string(self):
        cmd = CommandAuthChallenge(nonce='A_REAL_NONCE')
        assert cmd.serialize() == 'AUTHCHALLENGE SAFECOOKIE "A_REAL_NONCE"\r\n'

    def test_drop_guards(self):
        cmd = CommandDropGuards()
        assert cmd.serialize() == 'DROPGUARDS\r\n'

    def test_hs_fetch(self):
        cmd = CommandHsFetch(address='facebookcorewwwi')
        assert cmd.serialize() == 'HSFETCH facebookcorewwwi\r\n'

    def test_hs_fetch_with_servers(self):
        address = 'facebookcorewwwi'
        server1 = '$b34a4ac3892e41c58709d9c51b3648620a7d5bfe~Test1'
        server2 = '$7b70bf914770f022e71a26cbf3d9519dc89f2a9a~Test2'
        cmd = CommandHsFetch(
            address=address,
            servers=[
                server1,
                server2,
            ],
        )
        assert cmd.serialize() == f'HSFETCH {address} SERVER={server1} SERVER={server2}\r\n'

    def test_add_onion(self):
        port = VirtualPortAdapter.validate_python('80,127.0.0.1:80')
        cmd = CommandAddOnion(
            key=OnionServiceNewKeyStruct(OnionServiceKeyType.ED25519_V3),
            ports=[port],
        )
        adapter = cmd.adapter()
        serial = adapter.dump_python(cmd)
        assert serial == 'ADD_ONION NEW:ED25519-V3 Port=80,127.0.0.1:80\r\n'

    def test_add_onion_bytes(self):
        key = Ed25519PrivateKey.from_private_bytes(secrets.token_bytes(32))
        expanded = TrEd25519PrivateKey().to_expanded_bytes(key)
        expected = b64encode(expanded).decode().rstrip('=')

        port = VirtualPortAdapter.validate_python('80,127.0.0.1:80')
        cmd = CommandAddOnion(key=key, ports=[port])
        assert cmd.serialize() == (f'ADD_ONION ED25519-V3:{expected} Port=80,127.0.0.1:80\r\n')

    @pytest.mark.parametrize(
        ('value', 'type_'),
        [
            ('NEW:BEST', OnionServiceNewKeyStruct),
            (
                (
                    'ED25519-V3:ECum/PYnCBIHwWWmn6AaO29uY4Eq/hDEz6pLUG'
                    'znA0P0ZZKoLzYbJyURXRs0GNUz5aon9y+I3x3GauWJEXymSA'
                ),
                OnionServiceKeyStruct,
            ),
        ],
    )
    def test_add_onion_from_struct(self, value, type_):
        struct = {'key': value}
        adapter = CommandAddOnion.adapter()
        cmd = adapter.validate_python(struct)
        assert isinstance(cmd, CommandAddOnion)
        assert isinstance(cmd.key, type_)

    def test_add_onion_with_client_auth(self):
        auth = HsDescClientAuthV2(
            name='John',
            cookie=HsDescAuthCookie(
                auth_type=HsDescAuthTypeInt.BASIC_AUTH,
                cookie=bytes.fromhex('1a6608bb410a91de47e9b9692058378d'),
            ),
        )
        port = VirtualPortAdapter.validate_python('80,127.0.0.1:80')
        cmd = CommandAddOnion(
            key=OnionServiceNewKeyStruct('BEST'),
            ports=[port],
            flags={OnionServiceFlags.DISCARD_PK},
            max_streams=2,
            client_auth=[auth],
        )
        assert cmd.serialize() == (
            'ADD_ONION NEW:BEST Flags=DiscardPK MaxStreams=2 Port=80,127.0.0.1:80 '
            'ClientAuth=John:GmYIu0EKkd5H6blpIFg3jQA=\r\n'
        )

    def test_add_onion_with_client_auth_v3(self):
        v3_auth_strings = [
            '5BPBXQOAZWPSSXFKOIXHZDRDA2AJT2SWS2GIQTISCFKGVBFWBBDQ====',
            'RC3BHJ6WTBQPRRSMV65XGCZVSYJQZNWBQI3LLFS73VP6NHSIAD2Q====',
        ]
        auth_v3 = []
        for key_b32 in v3_auth_strings:
            auth_v3.append(X25519PublicKey.from_public_bytes(b32decode(key_b32)))

        port = VirtualPortAdapter.validate_python('80,127.0.0.1:80')
        cmd = CommandAddOnion(
            flags={OnionServiceFlags.V3AUTH},
            key=OnionServiceNewKeyStruct('BEST'),
            ports=[port],
            client_auth_v3=auth_v3,
        )
        assert cmd.serialize() == (
            'ADD_ONION NEW:BEST Flags=V3Auth Port=80,127.0.0.1:80 '
            'ClientAuthV3=5BPBXQOAZWPSSXFKOIXHZDRDA2AJT2SWS2GIQTISCFKGVBFWBBDQ '
            'ClientAuthV3=RC3BHJ6WTBQPRRSMV65XGCZVSYJQZNWBQI3LLFS73VP6NHSIAD2Q\r\n'
        )

    def test_add_onion_key_no_port(self):
        cmd = CommandAddOnion(key=OnionServiceNewKeyStruct('BEST'))
        msg = 'You must specify one or more virtual ports'
        with pytest.raises(PydanticSerializationError, match=msg):
            cmd.serialize()

    def test_del_onion(self):
        cmd = CommandDelOnion(address='facebookcorewwwi')
        assert cmd.serialize() == 'DEL_ONION facebookcorewwwi\r\n'

    def test_hs_post(self):
        cmd = CommandHsPost(descriptor='desc')
        assert cmd.serialize() == '+HSPOST\r\ndesc\r\n.\r\n'

    def test_hs_post_with_server(self):
        cmd = CommandHsPost(
            servers=['9695DFC35FFEB861329B9F1AB04C46397020CE31'],
            address='facebookcorewwwi',
            descriptor='desc',
        )
        assert cmd.serialize() == (
            '+HSPOST SERVER=9695DFC35FFEB861329B9F1AB04C46397020CE31 '
            'HSADDRESS=facebookcorewwwi\r\n'
            'desc\r\n.\r\n'
        )

    def test_onion_client_auth_add(self):
        address = 'aiostem26gcjyybsi3tyek6txlivvlc5tczytz52h4srsttknvd5s3qd'
        k64 = 'yPGUxgKaC5ACyEzsdANHJEJzt5DIqDRBlAFaAWWQn0o'
        key = X25519PrivateKey.from_private_bytes(b64decode(k64 + '='))
        cmd = CommandOnionClientAuthAdd(address=address, key=key)
        assert cmd.serialize() == f'ONION_CLIENT_AUTH_ADD {address} x25519:{k64}\r\n'

    def test_onion_client_auth_add_from_struct(self):
        address = 'aiostem26gcjyybsi3tyek6txlivvlc5tczytz52h4srsttknvd5s3qd'
        k64 = 'yPGUxgKaC5ACyEzsdANHJEJzt5DIqDRBlAFaAWWQn0o'
        adapter = CommandOnionClientAuthAdd.adapter()
        cmd = adapter.validate_python({'address': address, 'key': f'x25519:{k64}'})
        assert cmd.serialize() == f'ONION_CLIENT_AUTH_ADD {address} x25519:{k64}\r\n'

    def test_onion_client_auth_add_advanced(self):
        address = 'aiostem26gcjyybsi3tyek6txlivvlc5tczytz52h4srsttknvd5s3qd'
        k64 = 'yPGUxgKaC5ACyEzsdANHJEJzt5DIqDRBlAFaAWWQn0o'
        key = X25519PrivateKey.from_private_bytes(b64decode(k64 + '='))
        cmd = CommandOnionClientAuthAdd(
            address=address,
            key=key,
            nickname='Peter',
            flags={OnionClientAuthFlags.PERMANENT},
        )
        assert cmd.serialize() == (
            f'ONION_CLIENT_AUTH_ADD {address} x25519:{k64} '
            'ClientName=Peter Flags=Permanent\r\n'
        )

    def test_onion_client_auth_remove(self):
        address = 'aiostem26gcjyybsi3tyek6txlivvlc5tczytz52h4srsttknvd5s3qd'
        cmd = CommandOnionClientAuthRemove(address=address)
        assert cmd.serialize() == f'ONION_CLIENT_AUTH_REMOVE {address}\r\n'

    def test_onion_client_auth_view(self):
        cmd = CommandOnionClientAuthView()
        assert cmd.serialize() == 'ONION_CLIENT_AUTH_VIEW\r\n'

    def test_onion_client_auth_view_with_address(self):
        address = 'aiostem26gcjyybsi3tyek6txlivvlc5tczytz52h4srsttknvd5s3qd'
        cmd = CommandOnionClientAuthView(address=address)
        assert cmd.serialize() == f'ONION_CLIENT_AUTH_VIEW {address}\r\n'

    def test_drop_ownership(self):
        cmd = CommandDropOwnership()
        assert cmd.serialize() == 'DROPOWNERSHIP\r\n'

    def test_drop_timeouts(self):
        cmd = CommandDropTimeouts()
        assert cmd.serialize() == 'DROPTIMEOUTS\r\n'



from __future__ import annotations

import base64
import gc
import hashlib
import logging
import weakref
from datetime import UTC, datetime, timedelta
from ipaddress import IPv4Address, IPv6Address

import pytest
import pytest_asyncio
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from pydantic import ValidationError

from aiostem.event import (
    EventBandwidth,
    EventCellStats,
    EventCirc,
    EventCircBW,
    EventCircMinor,
    EventClientsSeen,
    EventConfChanged,
    EventConnBW,
    EventDescChanged,
    EventDisconnect,
    EventGuard,
    EventHsDesc,
    EventHsDescContent,
    EventNewConsensus,
    EventNewDesc,
    EventOrConn,
    EventSignal,
    EventStream,
    EventStreamBW,
    EventTbEmpty,
    EventUnknown,
    EventWord,
    EventWordInternal,
    event_from_message,
)
from aiostem.exceptions import CryptographyError, MessageError, ReplySyntaxError
from aiostem.structures import (
    CircuitEvent,
    HiddenServiceAddressV3,
    HsDescAction,
    HsDescFailReason,
    HsDescV2,
    HsDescV3,
    LogSeverity,
    LongServerName,
    Signal,
    StatusActionGeneral,
)
from aiostem.utils import MessageData, TrRSAPublicKey

from .test_reply import create_message

# All test coroutines will be treated as marked for asyncio.
pytestmark = pytest.mark.asyncio


class TestEvents:
    """Check that events are properly parsed."""

    async def test_error_message_not_event(self):
        message = await create_message(['250 OK'])
        with pytest.raises(MessageError, match='The provided message is not an event!'):
            event_from_message(message)

    async def test_disconnect(self):
        line = '650 DISCONNECT'
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventDisconnect)
        assert event.TYPE == EventWordInternal.DISCONNECT

    async def test_unknown(self):
        line = '650 UNKNOWN "This is a weird message"'
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventUnknown)
        assert event.message == message
        assert event.TYPE is None

    async def test_desc_changed(self):
        message = await create_message(['650 DESCCHANGED'])
        event = event_from_message(message)
        assert isinstance(event, EventDescChanged)

    async def test_guard(self):
        line = (
            '650 GUARD ENTRY $669E9D3CF2C1BF3A9E7A0B7FD89F8B4B5E1EF516~PalestineWillBeFree NEW'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventGuard)
        assert event.name.nickname == 'PalestineWillBeFree'
        assert event.type == 'ENTRY'
        assert event.status == 'NEW'

    async def test_clients_seen(self):
        line = (
            '650 CLIENTS_SEEN TimeStarted="2008-12-25 23:50:43" '
            'CountrySummary=us=16,de=8,uk=8 '
            'IPVersions=v4=16,v6=40'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventClientsSeen)
        assert isinstance(event.time, datetime)
        assert len(event.countries) == 3
        assert len(event.ip_versions) == 2
        assert event.countries['us'] == 16
        assert event.ip_versions['v6'] == 40

    async def test_stream_bw(self):
        line = '650 STREAM_BW 207187 20600 0 2025-01-04T21:29:50.985239'
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventStreamBW)
        assert isinstance(event.time, datetime)
        assert event.stream == 207187
        assert event.written == 20600
        assert event.read == 0

    @pytest.mark.parametrize('verb', ['NEWCONSENSUS', 'NS'])
    async def test_consensus_no_data(self, verb):
        message = await create_message([f'650 {verb}'])
        with pytest.raises(ReplySyntaxError, match='has no data attached to it'):
            event_from_message(message)

    async def test_new_consensus_with_data(self):
        lines = [
            '650+NEWCONSENSUS',
            (
                'r eisbaer AGHSKv0fBtTm81AGvT2cIdeYHqk W3YydCbm7DgcoHQxlxbjmJoCpWA '
                '2038-01-01 00:00:00 109.70.100.70 9002 0'
            ),
            'a [2a03:e600:100::70]:9002',
            's Exit Fast Running Stable Valid',
            'w Bandwidth=40000',
            (
                'r WishMaster ASPb4bj4Hgs1B5TP5K4XLhL9pAE 3z7CNbMYDiOzzHUoHc7hjUtkJc4 '
                '2038-01-01 00:00:00 31.192.107.132 443 0'
            ),
            's Fast Running Stable V2Dir Valid',
            'w Bandwidth=1100',
            (
                'r artikel10ber116 /9C5SdBnbxB2uySP5ABl0tpGVQk ZthfzyM7LX7XdlCW7gJvxZJ6O6U '
                '2038-01-01 00:00:00 185.220.101.29 9004 0'
            ),
            '',  # This extra new-line is on purpose to make sure we skip it!
            'a [2a0b:f4c2::29]:9004',
            's Exit Fast Running Stable Valid',
            'w Bandwidth=44000',
            'p accept 20-21,80,443,853,873,989-990,1194,2086,3690,5222',
            '.',
            '650 OK',
        ]
        message = await create_message(lines)
        event = event_from_message(message)
        assert isinstance(event, EventNewConsensus)
        assert len(event.routers) == 3
        assert event.routers[0].bandwidth == 40000
        assert len(event.routers[0].flags) == 5
        assert len(event.routers[0].addresses) == 1
        assert event.routers[0].port_policy is None
        assert event.routers[0].dir_port is None
        assert event.routers[2].port_policy.policy == 'accept'
        assert len(event.routers[2].port_policy.ports) == 10

    async def test_conf_changed_empty(self):
        lines = ['650-CONF_CHANGED', '650 OK']
        message = await create_message(lines)
        event = event_from_message(message)
        assert isinstance(event, EventConfChanged)
        assert len(event) == 0

    async def test_conf_changed(self):
        lines = [
            '650-CONF_CHANGED',
            '650-SocksPort=unix:/run/tor/socks WorldWritable ExtendedErrors RelaxDirModeCheck',
            '650-SocksPort=0.0.0.0:9050',
            '650 OK',
        ]
        message = await create_message(lines)
        event = event_from_message(message)
        assert isinstance(event, EventConfChanged)
        assert len(event) == 1
        assert len(event['SocksPort']) == 2
        assert event['SocksPort'][1] == '0.0.0.0:9050'

    async def test_circ_minor(self):
        line = (
            '650 CIRC_MINOR 261373 PURPOSE_CHANGED '
            '$E6FA2613FA6325F83FAFF643C315DA725BC372B9~NowhereDOTmoe,'
            '$9E42073401F6B4046983C9AA027DDB1538587B93~ellenatfims,'
            '$76BACC90CBA71714918554156CAABE955E7A940F~Quetzalcoatl '
            'BUILD_FLAGS=NEED_CAPACITY,NEED_UPTIME PURPOSE=CONFLUX_LINKED '
            'TIME_CREATED=2025-01-04T17:52:34.407938 OLD_PURPOSE=CONFLUX_UNLINKED'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventCircMinor)
        assert isinstance(event.time_created, datetime)
        assert event.event == CircuitEvent.PURPOSE_CHANGED
        assert len(event.path) == 3

    async def test_circ_minor_no_path(self):
        line = (
            '650 CIRC_MINOR 27329 PURPOSE_CHANGED '
            'BUILD_FLAGS=NEED_CAPACITY,NEED_UPTIME PURPOSE=MEASURE_TIMEOUT '
            'TIME_CREATED=2025-01-16T17:55:02.368535 OLD_PURPOSE=CONFLUX_UNLINKED'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventCircMinor)
        assert event.path is None

    async def test_conn_bw(self):
        line = '650 CONN_BW ID=123 TYPE=EXIT READ=234 WRITTEN=345'
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventConnBW)
        assert event.conn_id == 123
        assert event.conn_type == 'EXIT'
        assert event.read == 234
        assert event.written == 345

    async def test_circ_bw(self):
        line = (
            '650 CIRC_BW ID=261239 READ=0 WRITTEN=509 TIME=2025-01-04T17:13:43.979647 '
            'DELIVERED_READ=0 OVERHEAD_READ=0 DELIVERED_WRITTEN=153 OVERHEAD_WRITTEN=345'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventCircBW)
        assert isinstance(event.time, datetime)
        assert event.read == 0
        assert event.written == 509
        assert event.written_delivered == 153
        assert event.written_overhead == 345
        assert event.slow_start is None

    async def test_cell_stats(self):
        line = (
            '650 CELL_STATS ID=14 OutboundQueue=19403 OutboundConn=15 '
            'OutboundAdded=create_fast:1,relay_early:2 '
            'OutboundRemoved=create_fast:1,relay_early:2 '
            'OutboundTime=create_fast:0,relay_early:10'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventCellStats)
        assert event.circuit == 14
        assert event.inbound_queue is None
        assert event.outbound_conn_id == 15
        assert event.outbound_queue == 19403
        assert event.outbound_added['create_fast'] == 1
        assert event.outbound_added['relay_early'] == 2
        assert event.outbound_time['create_fast'].microseconds == 0
        assert event.outbound_time['relay_early'].microseconds == 10000

    async def test_tb_empty_global(self):
        line = '650 TB_EMPTY GLOBAL READ=93 WRITTEN=92 LAST=100'
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventTbEmpty)
        assert event.bucket == 'GLOBAL'
        assert event.conn_id is None
        assert event.last.microseconds == 100000
        assert event.read.microseconds == 93000
        assert event.written.microseconds == 92000

    async def test_tb_empty_orconn(self):
        line = '650 TB_EMPTY ORCONN ID=16 READ=0 WRITTEN=0 LAST=100'
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventTbEmpty)
        assert event.bucket == 'ORCONN'
        assert event.conn_id == 16
        assert event.last.microseconds == 100000
        assert event.read.microseconds == 0
        assert event.written.microseconds == 0

    async def test_circuit_launched(self):
        line = (
            '650 CIRC 267807 LAUNCHED BUILD_FLAGS=IS_INTERNAL,NEED_CAPACITY,NEED_UPTIME '
            'PURPOSE=HS_VANGUARDS TIME_CREATED=2025-01-04T23:55:46.318138'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventCirc)
        assert event.circuit == 267807
        assert event.status == 'LAUNCHED'
        assert event.purpose == 'HS_VANGUARDS'
        assert len(event.build_flags) == 3
        assert isinstance(event.time_created, datetime)

    async def test_circuit_hs_pow(self):
        line = (
            '650 CIRC 267807 LAUNCHED BUILD_FLAGS=IS_INTERNAL,NEED_CAPACITY,NEED_UPTIME '
            'PURPOSE=HS_VANGUARDS HS_POW=v1,2 TIME_CREATED=2025-01-04T23:55:46.318138'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventCirc)
        assert event.hs_pow.type == 'v1'
        assert event.hs_pow.effort == 2

    async def test_circuit_closed(self):
        line = (
            '650 CIRC 288979 CLOSED '
            '$F50CF02A0E6A9D9B25F7EB220FC26F7BD1B74999~flowjob02,'
            '$1814DBD5E7E1839CA3F43847683A08F3BF3D5C91~isodiapher,'
            '$745732BEBB6CD9344B6828481229F14F19E1C464~Unfixed1,'
            '$1214952C0357904505AAE425D48D84131064E9AC~cakesnwaffles '
            'BUILD_FLAGS=IS_INTERNAL,NEED_CAPACITY,NEED_UPTIME PURPOSE=HS_CLIENT_INTRO '
            'HS_STATE=HSCI_DONE '
            'REND_QUERY=facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid '
            'TIME_CREATED=2025-01-05T21:54:19.138927 REASON=FINISHED'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventCirc)
        assert event.circuit == 288979
        assert event.status == 'CLOSED'
        assert event.rend_query == 'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid'
        assert event.reason == 'FINISHED'
        assert len(event.path) == 4

    async def test_stream_new(self):
        line = (
            '650 STREAM 210420 NEW 0 1.1.1.1:53 SOURCE_ADDR=172.18.0.1:54640 '
            'PURPOSE=USER CLIENT_PROTOCOL=SOCKS4 NYM_EPOCH=16 SESSION_GROUP=-11 '
            'ISO_FIELDS=SOCKS_USERNAME,SOCKS_PASSWORD,CLIENTADDR,SESSION_GROUP,NYM_EPOCH'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventStream)
        assert event.status == 'NEW'
        assert event.stream == 210420
        assert event.target.host == IPv4Address('1.1.1.1')
        assert event.target.node is None
        assert event.target.port == 53
        assert event.nym_epoch == 16
        assert event.session_group == -11
        assert len(event.iso_fields) == 5

    async def test_stream_closed(self):
        line = (
            '650 STREAM 210427 CLOSED 282747 '
            '94.16.31.131.$0B4190C676FAAD34EB2DCB9A288939476CEBCF32.exit:443 '
            'REASON=END REMOTE_REASON=DONE CLIENT_PROTOCOL=UNKNOWN NYM_EPOCH=0 '
            'SESSION_GROUP=-2 ISO_FIELDS=SESSION_GROUP'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventStream)
        assert event.status == 'CLOSED'
        assert event.stream == 210427
        assert event.circuit == 282747
        assert event.target.host == IPv4Address('94.16.31.131')
        assert event.target.node is not None
        assert event.target.port == 443
        assert event.nym_epoch == 0
        assert event.session_group == -2
        assert len(event.iso_fields) == 1

    async def test_or_conn(self):
        line = (
            '650 ORCONN '
            '$427526EBD012CFE50FCFFCBFE221EFFE199AFA8C~portugesecartel '
            'CLOSED REASON=DONE ID=205906'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventOrConn)
        assert event.status == 'CLOSED'
        assert event.reason == 'DONE'
        assert event.conn_id == 205906

    async def test_bandwidth(self):
        message = await create_message(['650 BW 1670343 1936996'])
        event = event_from_message(message)
        assert isinstance(event, EventBandwidth)
        assert event.read == 1670343
        assert event.written == 1936996

    async def test_new_desc(self):
        line = (
            '650 NEWDESC '
            '$F5B58FEE44573C3BFD7D176D918BA5B4057519D7~bistrv1 '
            '$14AE2154A26F1D42C3C3BEDC10D05FDD9F8545BB~freeasf'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventNewDesc)
        assert len(event.servers) == 2
        for server in event.servers:
            assert isinstance(server, LongServerName)

    async def test_hs_desc_minimal(self):
        line = (
            '650 HS_DESC REQUESTED facebookcorewwwi NO_AUTH '
            '$F5B58FEE44573C3BFD7D176D918BA5B4057519D7~bistrv1 '
            '6wn4xyr3l2m6g5z3dcnvygul2tozaxli'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventHsDesc)
        assert event.reason is None

    async def test_hs_desc_advanced(self):
        line = (
            '650 HS_DESC FAILED oftestt7ffa4tt7et5wab7xhnzeooavy2xdmn6dtfa4pot7dk4xhviid '
            'NO_AUTH $14AE2154A26F1D42C3C3BEDC10D05FDD9F8545BB~freeasf '
            'NHN9fUdcd/9nJF6PSF6/IzdqkCiEoCsexfMv+7SGpCQ REASON=NOT_FOUND'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventHsDesc)
        assert event.action == HsDescAction.FAILED
        assert event.reason == HsDescFailReason.NOT_FOUND

    async def test_hs_desc_content(self):
        lines = [
            '650+HS_DESC_CONTENT facebookcorewwwi 6wn4xyr3l2m6g5z3dcnvygul2tozaxli '
            '$F5B58FEE44573C3BFD7D176D918BA5B4057519D7~bistrv1',
            'STUFFF',
            '.',
            '650 OK',
        ]
        message = await create_message(lines)
        event = event_from_message(message)
        assert isinstance(event, EventHsDescContent)
        assert event.address == 'facebookcorewwwi'
        assert event.descriptor_text == 'STUFFF'

    async def test_hs_desc_content_invalid_syntax(self):
        lines = [
            '650 HS_DESC_CONTENT facebookcorewwwi 6wn4xyr3l2m6g5z3dcnvygul2tozaxli '
            '$F5B58FEE44573C3BFD7D176D918BA5B4057519D7~bistrv1',
        ]
        message = await create_message(lines)
        with pytest.raises(ReplySyntaxError, match="Event 'HS_DESC_CONTENT' has no data"):
            event_from_message(message)

    async def test_log_message_line(self):
        line = '650 DEBUG conn_write_callback(): socket 14 wants to write.'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.message == 'conn_write_callback(): socket 14 wants to write.'
        assert event.severity == LogSeverity.DEBUG

    async def test_log_message_as_data(self):
        lines = [
            '650+WARN',
            'THIS IS A WARNING',
            '> BE WARNED!',
            '.',
            '650 OK',
        ]
        message = await create_message(lines)
        event = event_from_message(message)
        assert event.message == 'THIS IS A WARNING\n> BE WARNED!'
        assert event.severity == LogSeverity.WARNING

    async def test_network_liveness(self):
        line = '650 NETWORK_LIVENESS UP'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.status == 'UP'
        assert bool(event.status) is True

    async def test_status_general_clock_jumped(self):
        line = '650 STATUS_GENERAL NOTICE CLOCK_JUMPED TIME=120'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.action == StatusActionGeneral.CLOCK_JUMPED
        assert isinstance(event.arguments.time, timedelta)
        assert int(event.arguments.time.total_seconds()) == 120

    async def test_status_general_clock_skew_with_ip(self):
        line = '650 STATUS_GENERAL NOTICE CLOCK_SKEW SKEW=120 SOURCE=OR:1.1.1.1:443'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.action == StatusActionGeneral.CLOCK_SKEW
        assert isinstance(event.arguments.skew, timedelta)
        assert event.arguments.source.name == 'OR'
        assert event.arguments.source.address.host == IPv4Address('1.1.1.1')
        assert event.arguments.source.address.port == 443

    async def test_status_general_clock_skew_with_consensus(self):
        line = '650 STATUS_GENERAL NOTICE CLOCK_SKEW SKEW=120 SOURCE=CONSENSUS'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.action == StatusActionGeneral.CLOCK_SKEW
        assert event.arguments.source.name == 'CONSENSUS'
        assert event.arguments.source.address is None

    async def test_status_general_dir_all_unreachable(self):
        line = '650 STATUS_GENERAL ERR DIR_ALL_UNREACHABLE'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.action == StatusActionGeneral.DIR_ALL_UNREACHABLE
        assert event.arguments is None

    async def test_status_general_unknown_action(self, caplog):
        line = '650 STATUS_GENERAL NOTICE UNKNOWN_ACTION ARG=VAL'
        message = await create_message([line])
        with (
            caplog.at_level(logging.INFO, logger='aiostem'),
            pytest.raises(ValidationError, match='1 validation error for EventStatusGeneral'),
        ):
            event_from_message(message)
        assert "No syntax handler for action 'UNKNOWN_ACTION'" in caplog.text

    async def test_status_client_bootstrap(self):
        line = '650 STATUS_CLIENT NOTICE BOOTSTRAP PROGRESS=100 TAG=done SUMMARY="Done"'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.arguments.progress == 100
        assert event.arguments.summary == 'Done'
        assert event.arguments.tag == 'done'

    async def test_status_client_dangerous_socks_ipv4(self):
        line = '650 STATUS_CLIENT WARN DANGEROUS_SOCKS PROTOCOL=SOCKS5 ADDRESS=1.1.1.1:53'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.arguments.address.host == IPv4Address('1.1.1.1')
        assert event.arguments.address.port == 53
        assert event.arguments.protocol == 'SOCKS5'

    async def test_status_client_dangerous_socks_ipv6(self):
        line = (
            '650 STATUS_CLIENT WARN DANGEROUS_SOCKS PROTOCOL=SOCKS5 '
            'ADDRESS=[2a04:fa87:fffd::c000:426c]:443'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert event.arguments.address.host == IPv6Address('2a04:fa87:fffd::c000:426c')
        assert event.arguments.address.port == 443
        assert event.arguments.protocol == 'SOCKS5'

    async def test_status_client_socks_bad_hostname(self):
        line = '650 STATUS_CLIENT WARN SOCKS_BAD_HOSTNAME HOSTNAME="google.exit"'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.arguments.hostname == 'google.exit'

    async def test_status_transport_launched(self):
        line = '650 TRANSPORT_LAUNCHED client obfs4 127.0.0.1 1234'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.side == 'client'
        assert event.port == 1234

    async def test_status_pt_log(self):
        line = (
            '650 PT_LOG PT=/usr/bin/obs4proxy SEVERITY=debug MESSAGE="Connected to bridge A"'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert event.program == '/usr/bin/obs4proxy'
        assert event.severity == LogSeverity.DEBUG
        assert event.message == 'Connected to bridge A'

    async def test_status_pt_status(self):
        line = (
            '650 PT_STATUS PT=/usr/bin/obs4proxy TRANSPORT=obfs4 '
            'ADDRESS=198.51.100.123:1234 CONNECT=Success'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert event.program == '/usr/bin/obs4proxy'
        assert event.transport == 'obfs4'
        assert event.values['ADDRESS'] == '198.51.100.123:1234'
        assert event.values['CONNECT'] == 'Success'

    async def test_addr_map_standard(self):
        line = (
            '650 ADDRMAP google.com 142.250.74.110 "2024-12-08 23:00:36" '
            'EXPIRES="2024-12-08 23:00:36" CACHED="NO" STREAMID=109038'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert event.original == 'google.com'
        assert event.replacement == IPv4Address('142.250.74.110')
        assert event.cached is False

    async def test_addr_map_error(self):
        line = (
            '650 ADDRMAP 2a04:fa87:fffd::c000:426c <error> "2024-12-09 07:24:03" '
            'error=yes EXPIRES="2024-12-09 07:24:03" CACHED="NO" STREAMID=110330'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert event.original == IPv6Address('2a04:fa87:fffd::c000:426c')
        assert event.replacement is None
        assert isinstance(event.expires, datetime)
        assert event.expires.tzinfo == UTC
        assert event.stream == 110330
        assert event.cached is False
        assert event.error == 'yes'

    async def test_addr_map_permanent(self):
        line = '650 ADDRMAP dns.google 8.8.8.8 NEVER CACHED="YES"'
        message = await create_message([line])
        event = event_from_message(message)
        assert event.expires is None
        assert event.cached is True

    async def test_build_timeout_set(self):
        line = (
            '650 BUILDTIMEOUT_SET COMPUTED TOTAL_TIMES=1000 TIMEOUT_MS=815 '
            'XM=283 ALPHA=1.520695 CUTOFF_QUANTILE=0.800000 TIMEOUT_RATE=0.292260 '
            'CLOSE_MS=60000 CLOSE_RATE=0.011098'
        )
        message = await create_message([line])
        event = event_from_message(message)
        assert event.total_times == 1000
        assert event.xm.microseconds == 283000

    async def test_signal(self):
        line = '650 SIGNAL RELOAD'
        message = await create_message([line])
        event = event_from_message(message)
        assert isinstance(event, EventSignal)
        assert event.signal == Signal.RELOAD
        assert event.TYPE == EventWord.SIGNAL


class TestHsDescriptors:
    """Check parsing of onion descriptors."""

    @pytest.fixture(scope='session')
    def hs_desc_v2_lines(self):
        address = 'facebookcorewwwi'
        path = f'tests/samples/{address}/content.txt'
        with open(path) as fp:
            return list(map(str.rstrip, fp))

    @pytest_asyncio.fixture(scope='function')
    async def hs_desc_v2_event(self, hs_desc_v2_lines):
        message = await create_message(hs_desc_v2_lines)
        return event_from_message(message)

    @pytest.fixture(scope='session')
    def hs_desc_v3_lines(self):
        address = 'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid'
        path = f'tests/samples/{address}/content.txt'
        with open(path) as fp:
            return list(map(str.rstrip, fp))

    @pytest_asyncio.fixture(scope='function')
    async def hs_desc_v3_event(self, hs_desc_v3_lines):
        message = await create_message(hs_desc_v3_lines)
        return event_from_message(message)

    @pytest.fixture(scope='session')
    def hs_desc_v3_auth_lines(self):
        address = 'vakp3s3f4f5uopzvzsczsumurwrawuc5sexvjxqgzng6untov3gc5lyd'
        path = f'tests/samples/{address}/content.txt'
        with open(path) as fp:
            return list(map(str.rstrip, fp))

    @pytest_asyncio.fixture(scope='function')
    async def hs_desc_v3_auth_event(self, hs_desc_v3_auth_lines):
        message = await create_message(hs_desc_v3_auth_lines)
        return event_from_message(message)

    async def test_hs_desc_v2(self, hs_desc_v2_event):
        event = hs_desc_v2_event
        assert isinstance(event, EventHsDescContent)
        assert event.address == 'facebookcorewwwi'
        assert isinstance(event.descriptor, HsDescV2)

        # Check general purpose fields from the descriptor.
        desc = event.descriptor
        assert desc.protocol_versions == {2, 3}
        assert desc.version == 2

        # Check that the key matches the domain name!
        pubkey = TrRSAPublicKey().to_bytes(desc.permanent_key)
        digest = hashlib.sha1(pubkey).digest()  # noqa: S324
        computed = base64.b32encode(digest[:10]).decode('ascii')
        assert computed.lower() == event.address

        # Parse and get a list of introduction points.
        intros = desc.introduction_points()
        assert len(intros) == 10

        first_intro = intros[0]
        assert first_intro.ip == IPv4Address('212.74.233.20')
        assert first_intro.onion_port == 9101

        desc.raise_for_invalid_signature()

    async def test_hs_desc_v2_signature_error(self, hs_desc_v2_event):
        assert isinstance(hs_desc_v2_event, EventHsDescContent)
        descriptor = hs_desc_v2_event.descriptor
        descriptor.signature = bytes.fromhex(
            '028da11bc35e2baff1ea80054370d471fe716d3fb3428299517225e1458d271e'
            '06cb8bcf94282b46bbbf728d834d637ccea3f0db42c0283a598d1d69beab0ad6'
            '4403fbed851d5e305d8be95c98d94c7547570041d3a0f8c57760d20e9122d537'
            'a283f18542df04aeb0acc6feac219285586d71f0d12e20bf3becd97e17ec1761'
        )
        msg = 'Decrypted certificate signature has an invalid format'
        with pytest.raises(CryptographyError, match=msg):
            descriptor.raise_for_invalid_signature()

    async def test_hs_desc_v2_intro_errors(self, hs_desc_v2_event):
        msg = 'Authentication cookie for V2 descriptor is not yet implemented'
        with pytest.raises(NotImplementedError, match=msg):
            hs_desc_v2_event.descriptor.introduction_points('password')

    async def test_hs_desc_v3(self, hs_desc_v3_event):
        address = 'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid'
        event = hs_desc_v3_event

        assert isinstance(event, EventHsDescContent)
        assert isinstance(event.descriptor, HsDescV3)
        assert event.address == address

        # Check general purpose fields from the descriptor.
        desc = event.descriptor
        assert desc.hs_descriptor == 3
        assert desc.revision == 2602216024
        assert int(desc.lifetime.total_seconds()) == 10800

        # Additional checks on the signing certificate.
        signing_cert = desc.signing_cert
        assert signing_cert.expired is True

        signing_key = signing_cert.signing_key
        assert signing_key is not None

        signing_cert.raise_for_invalid_signature(signing_key)
        desc.raise_for_invalid_signature()

        layer1 = desc.decrypt_layer1(event.address)
        assert len(layer1.auth_clients) == 16

        layer2 = desc.decrypt_layer2(event.address)
        assert layer2.single_service is True

        # Check introduction points.
        assert len(layer2.introduction_points) == 10
        for intro in layer2.introduction_points:
            intro.auth_key_cert.raise_for_invalid_signature(intro.auth_key_cert.signing_key)
            intro.enc_key_cert.raise_for_invalid_signature(intro.enc_key_cert.signing_key)

        # Check a random link specifier in our structure.
        links = layer2.introduction_points[2].link_specifiers
        assert links[0].host == IPv4Address('193.142.147.204')
        assert links[0].port == 9200

    async def test_hs_desc_v3_cache(self, hs_desc_v3_auth_event, hs_desc_v3_event):
        desc1 = HsDescV3.from_text(hs_desc_v3_event.descriptor_text)
        desc2 = HsDescV3.from_text(hs_desc_v3_auth_event.descriptor_text)
        info1 = desc1.decrypt_layer1.cache_info()
        info2 = desc2.decrypt_layer1.cache_info()
        assert info1.currsize == 0
        assert info2.currsize == 0

        for _ in range(2):
            desc1.decrypt_layer1(hs_desc_v3_event.address)
            desc2.decrypt_layer1(hs_desc_v3_auth_event.address)
            info1 = desc1.decrypt_layer1.cache_info()
            info2 = desc2.decrypt_layer1.cache_info()
            assert info1.currsize == 1
            assert info2.currsize == 1

    async def test_hs_desc_v3_cache_gc(self, hs_desc_v3_event):
        """Check that the descriptor is properly garbage collected."""
        desc = HsDescV3.from_text(hs_desc_v3_event.descriptor_text)
        desc.decrypt_layer2(hs_desc_v3_event.address)
        desc_ref = weakref.ref(desc)
        assert desc_ref() == desc

        desc = None
        gc.collect()
        assert desc_ref() is None

    async def test_hs_desc_v3_auth_no_key(self, hs_desc_v3_auth_event):
        """Check with an encrypted descriptor but with no key."""
        desc = hs_desc_v3_auth_event.descriptor
        assert desc.hs_descriptor == 3
        assert desc.revision == 2634456021

        layer1 = desc.decrypt_layer1(hs_desc_v3_auth_event.address)
        assert len(layer1.auth_clients) == 16

        with pytest.raises(CryptographyError, match='Invalid MAC, something is corrupted!'):
            desc.decrypt_layer2(hs_desc_v3_auth_event.address)

    async def test_hs_desc_v3_no_signature(self, hs_desc_v3_lines):
        message = await create_message(hs_desc_v3_lines)

        # Alter the input message to filter out the signature part (at the end).
        lines = []
        for line in message.items[0].data.splitlines():
            if line.startswith('signature '):
                break
            lines.append(line)

        # Replace the message data by a new message we are creating.
        msgdata = MessageData(
            status=message.items[0].status,
            header=message.items[0].header,
            data='\n'.join(lines),
        )
        message.items.clear()
        message.items.append(msgdata)

        # Check that we get an error because there is no signature.
        event = event_from_message(message)
        msg = 'No signature found on the HsV3 descriptor'
        with pytest.raises(ReplySyntaxError, match=msg):
            event.descriptor  # noqa: B018

    async def test_hs_desc_v3_signature_error(self, hs_desc_v3_event):
        hs_desc_v3_event.descriptor.signature = b'abcdf'
        with pytest.raises(CryptographyError, match='Descriptor has an invalid signature'):
            hs_desc_v3_event.descriptor.raise_for_invalid_signature()

    async def test_hs_desc_v3_no_signing_cert_key(self, hs_desc_v3_event):
        # Clear all extensions (including the one holding the signing key).
        hs_desc_v3_event.descriptor.signing_cert.key = None
        hs_desc_v3_event.descriptor.raise_for_invalid_signature()

    async def test_hs_desc_v3_cert_no_signing_key(self, hs_desc_v3_event):
        # Clear all extensions (including the one holding the signing key).
        hs_desc_v3_event.descriptor.signing_cert.extensions.clear()
        with pytest.raises(ReplySyntaxError, match='No signing key found in the descriptor'):
            hs_desc_v3_event.descriptor.decrypt_layer1(hs_desc_v3_event.address)

    async def test_hs_desc_v3_bad_decrypt_addr(self, hs_desc_v3_event):
        addr_str = '2gzyxa5ihm7nsggfxnu52rck2vv4rvmdlkiu3zzui5du4xyclen53wid.onion'
        address = HiddenServiceAddressV3.from_string(addr_str)
        with pytest.raises(CryptographyError, match='Invalid MAC, something is corrupted!'):
            hs_desc_v3_event.descriptor.decrypt_layer1(address)

    async def test_hs_desc_v3_auth_with_valid_key(self, hs_desc_v3_auth_event):
        """Check with an encrypted descriptor but with a valid key."""
        hexkey = 'D8D9ABB522BBC0A2E35CCDDD6B4CECF2FB4D51F7698EC8D1E0F2A1D81D340950'
        key = X25519PrivateKey.from_private_bytes(bytes.fromhex(hexkey))
        desc = hs_desc_v3_auth_event.descriptor
        layer2 = desc.decrypt_layer2(hs_desc_v3_auth_event.address, key)
        assert len(layer2.introduction_points) == 3
        assert layer2.single_service is False

    async def test_hs_desc_v3_auth_with_invalid_key(self, hs_desc_v3_event):
        addr_str = 'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid.onion'
        address = HiddenServiceAddressV3.from_string(addr_str)
        msg = 'No client matching the secret key was found in the descriptor'
        with pytest.raises(CryptographyError, match=msg):
            hs_desc_v3_event.descriptor.decrypt_layer2(address, X25519PrivateKey.generate())



from __future__ import annotations

import logging
from collections import OrderedDict
from collections.abc import (
    Iterable,
    Mapping,
    Sequence,
    Set as AbstractSet,
)
from dataclasses import dataclass, field
from enum import IntFlag
from typing import TYPE_CHECKING

from ..exceptions import ReplySyntaxError
from .message import MessageData, MessageLine

if TYPE_CHECKING:
    from .message import BaseMessage

logger = logging.getLogger(__package__)


def _string_indexof(string: str, separators: str) -> int:
    """
    Find the index of any of the provided separator.

    Args:
        string: A string to look into.
        separators: A list of possible separators.

    Returns:
        The index of the found separator.

    """
    idx = len(string)
    for i, c in enumerate(string):
        if c in separators:
            idx = i
            break
    return idx


def _string_unescape(string: str) -> tuple[str, str]:
    """
    Unescape the provided quoted string.

    Args:
        string: The string to unescape, starting with `"`.

    Raises:
        ReplySyntaxError: ``EOF`` was reached before the closing quote.

    Returns:
        A tuple with the parsed value and the remaining string.

    """
    escaping = False
    result = ''
    index = None  # type: int | None

    for i in range(1, len(string)):
        c = string[i]
        if escaping:
            result += c
            escaping = False
        elif c == '\\':
            escaping = True
        elif c == '"':
            index = i
            break
        else:
            result += c

    if index is None:
        msg = 'No double-quote found before the end of string.'
        raise ReplySyntaxError(msg)

    return result, string[index + 1 :]


class ReplySyntaxFlag(IntFlag):
    """All accepted flags for a reply syntax item."""

    #: Capture everything remaining in the last positional argument.
    POS_REMAIN = 1
    #: Enable the parsing of keyword arguments.
    KW_ENABLE = 16
    #: Whether the keyword value can be enclosed with quotes.
    KW_QUOTED = 32
    #: Whether we preserve unknown keyword arguments.
    KW_EXTRA = 64
    #: Keyword arguments can omit their key.
    KW_OMIT_KEYS = 128
    #: Keyword arguments can omit their value.
    KW_OMIT_VALS = 256
    #: Use data from :class:`.MessageData` as a potential KW value.
    KW_USE_DATA = 512
    #: No quoting or escape is performed (whole line is a value).
    KW_RAW = 1024


@dataclass(kw_only=True, slots=True)
class ReplySyntax:
    """
    Describe the syntax of a single reply item.

    Important:
        - :attr:`args_max` is set to ``max(args_max, len(args_map))``.
        - :attr:`args_min` cannot be greater than :attr:`args_max`.
        - :attr:`kwargs_map` must be empty when :data:`~ReplySyntaxFlag.KW_ENABLE` is not set.
        - :attr:`args_min` must be equal to :attr:`args_max` when
          :data:`~ReplySyntaxFlag.KW_ENABLE` is set.
        - :data:`~ReplySyntaxFlag.POS_REMAIN` is mutually exclusive with
          :data:`~ReplySyntaxFlag.KW_ENABLE`.
        - :data:`~ReplySyntaxFlag.KW_QUOTED` is mutually exclusive with
          :data:`~ReplySyntaxFlag.KW_RAW`.

    """

    #: Minimum number of required positional arguments.
    args_min: int = 0
    #: Maximum number of positional arguments.
    args_max: int = 0
    #: List of names for the positional arguments (:obj:`None` to ignore it).
    args_map: Sequence[str | None] = field(default_factory=list)
    #: Correspondance map for keyword arguments.
    kwargs_map: Mapping[str | None, str] = field(default_factory=dict)
    #: These KW mapping keys can hold multiple values.
    kwargs_multi: AbstractSet[str] = field(default_factory=frozenset)
    #: List of parsing flags.
    flags: ReplySyntaxFlag = field(default_factory=lambda: ReplySyntaxFlag(0))

    def __post_init__(self) -> None:
        """Check syntax incompatibilities."""
        args_max = max(self.args_max, len(self.args_map))
        if self.args_min > args_max:
            msg = 'Minimum argument count is greater than the maximum.'
            raise RuntimeError(msg)

        if len(self.kwargs_map) and not (self.flags & ReplySyntaxFlag.KW_ENABLE):
            msg = 'Keywords are disabled but we found items in its map.'
            raise RuntimeError(msg)

        if self.args_min < args_max and (self.flags & ReplySyntaxFlag.KW_ENABLE):
            msg = 'Cannot have optional argument along with keyword arguments.'
            raise RuntimeError(msg)

        # Cannot capture positional remains and enable KW flags
        remain_vs_kw = ReplySyntaxFlag.POS_REMAIN | ReplySyntaxFlag.KW_ENABLE
        if self.flags & remain_vs_kw == remain_vs_kw:
            msg = 'Positional remain and keywords are mutually exclusive.'
            raise RuntimeError(msg)

        # Cannot both omit keys and values.
        omit_keys_vals = ReplySyntaxFlag.KW_OMIT_KEYS | ReplySyntaxFlag.KW_OMIT_VALS
        if self.flags & omit_keys_vals == omit_keys_vals:
            msg = 'KW_OMIT_KEYS and KW_OMIT_VALS are mutually exclusive.'
            raise RuntimeError(msg)

        # Cannot both handle quotes and raw values.
        raw_vs_quoted = ReplySyntaxFlag.KW_RAW | ReplySyntaxFlag.KW_QUOTED
        if self.flags & raw_vs_quoted == raw_vs_quoted:
            msg = 'KW_RAW and KW_QUOTED are mutually exclusive.'
            raise RuntimeError(msg)

        self.args_max = args_max

    def _iter_keywords(
        self,
        string: str,
        data: str | None = None,
    ) -> Iterable[tuple[str | None, str | None]]:
        """
        Iterate the string to extract key and value pairs.

        Note:
            Returned key and value cannot be both set to :obj:`None`.

        Args:
            string: The input string to parse from.
            data: Optional data blob (to used with `KW_USE_DATA`).

        Raises:
            ReplySyntaxError: when the key/value syntax is invalid.

        Yields:
            Pairs of key and values.

        """
        while len(string):
            key = None  # type: str | None
            val = None  # type: str | None
            omit_vals = False

            # Remove any leading space of any kind.
            string = string.lstrip(' \t\r\v\n')

            if len(string) and string[0] != '"':
                idx = _string_indexof(string, ' \t\r\v\n=')
                if idx < len(string) and string[idx] == '=':
                    key = string[:idx]
                    string = string[idx + 1 :]
                elif self.flags & ReplySyntaxFlag.KW_OMIT_VALS:
                    key = string[:idx]
                    string = string[idx:]
                    omit_vals = True

            if key is None and not (self.flags & ReplySyntaxFlag.KW_OMIT_KEYS):
                msg = 'Got a single string without either OMIT_KEYS or OMIT_VALS.'
                raise ReplySyntaxError(msg)

            if not omit_vals:
                if len(string):
                    if self.flags & ReplySyntaxFlag.KW_RAW:
                        val = string
                        string = ''
                    elif string[0] == '"':
                        if not (self.flags & ReplySyntaxFlag.KW_QUOTED):
                            msg = 'Got an unexpected quoted value.'
                            raise ReplySyntaxError(msg)

                        val, string = _string_unescape(string)
                    else:
                        idx = _string_indexof(string, ' \t\r\v\n')
                        val = string[:idx]
                        string = string[idx:]
                elif data is not None and self.flags & ReplySyntaxFlag.KW_USE_DATA:
                    val = data
                else:
                    val = ''

            yield (key, val)

    def parse(
        self,
        message: BaseMessage,
    ) -> Mapping[str, Sequence[str | None] | str | None]:
        """
        Parse the provided message.

        Args:
            message: A message or sub-message to parse with this syntax.

        Returns:
            A map of parsed values.

        """
        # Capture remain as a part of the argument list if told to do so.
        do_remain = bool(self.flags & ReplySyntaxFlag.POS_REMAIN)
        items = message.header.split(' ', maxsplit=self.args_max - int(do_remain))
        if len(items) < self.args_min:
            msg = 'Received too few arguments on the reply.'
            raise ReplySyntaxError(msg)

        result = OrderedDict()  # type: OrderedDict[str, list[str | None] | str | None]
        for i in range(min(len(items), len(self.args_map))):
            key = self.args_map[i]
            if key is not None:
                result[key] = items[i]

        if len(items) > self.args_max:
            remain = items[self.args_max]
            if not (self.flags & ReplySyntaxFlag.KW_ENABLE):
                msg = f"Unexpectedly found remaining data: '{remain}'"
                raise ReplySyntaxError(msg)

            data = message.data if isinstance(message, MessageData) else None
            for original_key, val in self._iter_keywords(remain, data):
                do_include = True
                has_key = original_key in self.kwargs_map
                if has_key:
                    key = self.kwargs_map[original_key]
                elif self.flags & ReplySyntaxFlag.KW_EXTRA:
                    key = original_key
                else:
                    do_include = False

                if do_include and key is not None:
                    if key in self.kwargs_multi:
                        existing = result.setdefault(key, [])
                        if isinstance(existing, list):  # pragma: no branch
                            existing.append(val)
                    else:
                        result[key] = val
                else:
                    logger.info(f'Found an unhandled keyword: {original_key}={val}')

        return result

    def parse_string(
        self,
        string: str,
    ) -> Mapping[str, Sequence[str | None] | str | None]:
        """
        Parse the provided string.

        Args:
            string: A plain header to parse as if it was a message header.

        Returns:
            A map of parsed values.

        """
        return self.parse(MessageLine(status=250, header=string))
